# frozen_string_literal: true

require 'active_record'

module Segmentor
  # Segment is a slice of data, based on a condition.
  class Segment < ActiveRecord::Base
    self.table_name = :segmentor_segments

    NO_REPEAT = 0

    has_many :receipts, class_name: '::Segmentor::Receipt', dependent: :destroy
    has_many :sessions, class_name: '::Segmentor::Session', dependent: :destroy
    has_one :source, class_name: '::Segmentor::Source', inverse_of: :segment, dependent: :destroy
    has_one :notifier_context, class_name: '::Segmentor::NotifierContext', inverse_of: :segment, dependent: :destroy

    accepts_nested_attributes_for :source
    accepts_nested_attributes_for :notifier_context

    class_attribute :_after_session_change

    def self.after_session_change(method_name)
      self._after_session_change = method_name
    end

    # generate_frequency is in days, should be between 0 and 365. 0 means no repeat
    validates :generate_frequency, numericality: { only_integer: true,
                                                   greater_than_or_equal_to: 0,
                                                   less_than_or_equal_to: 365 }
    # generate_frequency is in days, should be between 0 and 365. 0 means no repeat
    validates :notify_frequency, numericality: { only_integer: true,
                                                 greater_than_or_equal_to: 0,
                                                 less_than_or_equal_to: 365 }

    validates :name, length: { minimum: 3, maximum: 255 }, presence: true
    validates :description, length: { maximum: 255 }

    # generate all the targets for this segment
    # it then returns the session
    def generate!
      # start a new session

      session = start_session
      begin
        # get the data from the source
        returned_targets = source.execute
        # replace the data in the segment in a transaction
        returned_targets.each do |target|
          target.session = session
          target.save!
        end

        # mark the session as draft
        end_session(session)

        activate(session) if auto_activate

        session
      rescue ::Segmentor::Errors::EvaluationError => e
        # mark the session as failed with a reason. don't raise
        session.mark_as_failed!(e.message)
        session
      rescue StandardError
        # delete the session and raise
        session.destroy
        raise
      end
    end

    # notify! calls the given block for each target that can be notified
    # if the block raises a Segmentor::Errors::NotificationError, the target is skipped
    # and no receipt is issued but the rest of the targets are still notified
    # as most notification systems do not support transactions, it is important
    # to ensure only ::Segmentor::Errors::NotificationError is raised unless you are
    # sure you want to rollback all notification receipts
    def notify!(&block)
      # get the active targets from the segment
      # for each target, check if its receipt is within segment repeat frequency
      # if not, send the notification and generate a receipt
      # if so, do nothing
      raise ::Segmentor::Errors::NoSessionError, 'no active session' unless active_session

      ::ActiveRecord::Base.transaction do
        active_session.targets.each do |target|
          next unless target.can_be_notified?

          # send the notification
          # generate a receipt
          result = block.call(target, notifier_context) if block_given?
          target.issue_receipt!(result)
        # catch ::Segmentor::Errors::NotificationError and do nothing
        rescue ::Segmentor::Errors::NotificationError
          # do nothing. so we only skip issue_receipt! but continue to the rest of the targets
          # any other error will be raised and rollback the transaction
        end
        active_session.touch
      end
    end

    # resets the last session of this segment to be in progress again
    def reset!
      latest_session.mark_as_draft!
    end

    def last_run
      # latest_session is going to be either active, draft or failed
      latest_session.try(:updated_at)
    end

    # turns a session from draft into active and marks
    # the current active session as archived.
    # if no current active session is found. the latest session is marked as active
    def activate(session)
      ::ActiveRecord::Base.transaction do
        # mark the current active session as archived
        as = active_session
        if as
          as.status = ::Segmentor::Session::ARCHIVED
          as.save!
        end

        # mark the new session as active
        session.status = ::Segmentor::Session::ACTIVE
        session.save!
      end
    end

    def active_session
      sessions.active.first
    end

    def active_session!
      as = active_session
      raise ::Segmentor::Errors::NoSessionError, 'no active session' unless as

      as
    end

    def latest_session(include_running: false)
      unaccepted_status = [::Segmentor::Session::ARCHIVED]
      unaccepted_status << ::Segmentor::Session::IN_PROGRESS unless include_running

      # pick the latest session that's not in progress or archived
      sessions.where
              .not(status: unaccepted_status)
              .order(created_at: :desc)
              .first
    end

    private

    def start_session
      session = ::Segmentor::Session.new(segment: self)
      session.save!

      session
    end

    def end_session(session)
      session.mark_as_draft!
    end
  end
end
