# frozen_string_literal: true

module Segmentor
  # Session is a single segmentation session. It can be
  # draft, in progress, or active.
  # at any point, there is only 1 active session per segment is allowed
  # while draft and in progress, there can be multiple sessions
  # in progress is when a source is still generating targets and they
  # are being saved.
  # draft is when a session has the targets, but they are not the ones
  # used for notification. This can be used to test a new source code for example.
  class Session < ActiveRecord::Base
    self.table_name = :segmentor_sessions

    IN_PROGRESS = 0 # default
    DRAFT = 1
    ACTIVE = 2
    ARCHIVED = 3
    FAILED = 4

    belongs_to :segment, class_name: '::Segmentor::Segment'
    has_many :targets, class_name: '::Segmentor::Target', dependent: :destroy

    validates :status, presence: true
    validates :status, inclusion: { in: [IN_PROGRESS, DRAFT, ACTIVE, ARCHIVED, FAILED] }
    validates :segment_id, presence: true
    validate :only_one_active_session_per_segment
    validates :reason, presence: true, if: :failed?

    scope :active, -> { where(status: ACTIVE) }
    scope :latest, -> { order(created_at: :desc).limit(1).first }

    before_create :generate_session_id
    after_commit :call_segment_callbacks

    def mark_as_active!
      # only draft sessions can be made active
      raise ::Segmentor::Errors::SessionStateError, 'only draft sessions can be made active' unless status == DRAFT

      self.status = ACTIVE
      save!
    end

    def mark_as_draft!
      # archived and failed sessions cannot be turned into draft
      if status == ARCHIVED || status == FAILED
        raise ::Segmentor::Errors::SessionStateError, 'failed and archived sessions cannot be turned into draft'
      end

      self.status = DRAFT
      self.reason = nil
      save!
    end

    def mark_as_archived!
      # only active and failed sessions can be archived
      unless status == ACTIVE || status == FAILED
        raise ::Segmentor::Errors::SessionStateError, 'only active and failed sessions can be archived'
      end

      self.status = ARCHIVED
      save!
    end

    def mark_as_failed!(reason)
      # archived sessions cannot be marked as failed
      raise ::Segmentor::Errors::SessionStateError, 'archived sessions cannot be marked as failed' if status == ARCHIVED

      self.status = FAILED
      self.reason = reason
      save!
    end

    private

    def only_one_active_session_per_segment
      # only one active session per segment is allowed
      if status == ACTIVE && ::Segmentor::Session.where(status: ACTIVE, segment_id: segment_id).count.positive?
        raise ::Segmentor::Errors::SessionStateError, 'only one active session per segment is allowed'
      end
    end

    def generate_session_id
      # generate a session id
      self.session_id ||= SecureRandom.uuid
    end

    def failed?
      status == FAILED
    end

    def call_segment_callbacks
      segment.send(segment.class._after_session_change, self) if segment.class._after_session_change.present?
    end

  end
end
